/* rotzoomer - creates a collage of rotated and scaled portions of the screen
 * Copyright (C) 2001 Claudio Matsuoka <claudio@helllabs.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */

/* 
  * ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 01/03
  */

/* plan9-related stuff */
#include <u.h>
#include <libc.h>
#include <draw.h>
#include <event.h>

/* screenshot-related stuff */
#include "getscr.h"	
static uchar *scrbuf;
static Image *scrimg;
static Image *bufimg;

/* other defines */
#define NULL nil
#define XPoint Point
#define NRAND nrand
#define LRAND lrand
#define random lrand
#define MAXRAND ((2<<31)-1)
#define MAX(a, b) (((a) > (b))?(a):(b))
#define MIN(a, b) (((a) < (b))?(a):(b))
#define RAND_MAX MAXRAND
#define ABS abs
Image *colors[256];
#define M_PI	PI
#define Bool int
#define True 1
#define False 0
#define CYCLES 50000

char *buttons[] = 
{
	"exit",
	0
};

Menu menu = 
{
	buttons
};
Mouse m;



/* end of plan9-related defines */

struct zoom_area {
	int w, h;		/* rectangle width and height */
	int inc1, inc2;		/* rotation and zoom angle increments */
	int dx, dy;		/* translation increments */
	int a1, a2;		/* rotation and zoom angular variables */
	int ox, oy;		/* origin in the background copy */
	int xx, yy;		/* left-upper corner position (* 256) */
	int x, y;		/* left-upper corner position */
	int ww, hh;		/* valid area to place left-upper corner */
	int n;			/* number of iteractions */
	int count;		/* current iteraction */
};


static int width, height;
static struct zoom_area **zoom_box = nil;
static int num_zoom = 2;
static int move = 0;
static int sweep = 0;
static int delay = 10;
static int anim = 1;


static void rotzoom (struct zoom_area *za)
{
	int x, y, c, s, zoom, z;
	int x2 = za->x + za->w - 1, y2 = za->y + za->h - 1;
	int ox = 0, oy = 0;
	Rectangle r;

	z = 8100 * sin (M_PI * za->a2 / 8192);
	zoom = 8192 + z;

	c = zoom * cos (M_PI * za->a1 / 8192);
	s = zoom * sin (M_PI * za->a1 / 8192);
	for (y = za->y; y <= y2; y++) {
		for (x = za->x; x <= x2; x++) {
			ox = (x * c + y * s) >> 13;
			oy = (-x * s + y * c) >> 13;

			while (ox < 0)
				ox += width;
			while (oy < 0)
				oy += height;
			while (ox >= width)
				ox -= width;
			while (oy >= height)
				oy -= height;

			//XPutPixel (buffer_map, x, y, XGetPixel (scrbuf, ox, oy));

			r = screen->r;
			r.min = addpt(r.min, Pt(x, y));
			r.max = addpt(r.min, Pt(2, 2));

			draw(screen, r, scrimg, nil, addpt(screen->r.min, Pt(ox, oy)));
		}
	}

	za->a1 += za->inc1;		/* Rotation angle */
	za->a1 &= 0x3fff;;

	za->a2 += za->inc2;		/* Zoom */
	za->a2 &= 0x3fff;

	za->ox = ox;			/* Save state for next iteration */
	za->oy = oy;

	za->count++;
}


static void reset_zoom (struct zoom_area *za)
{
	if (sweep) {
		int speed = random () % 100 + 100;
		switch (random () % 4) {
		case 0:
			za->w = width;
			za->h = 10;
			za->x = 0;
			za->y = 0;
			za->dx = 0;
			za->dy = speed;
			za->n = (height - 10) * 256 / speed;
			break;
		case 1:
			za->w = 10;
			za->h = height;
			za->x = width - 10;
			za->y = 0;
			za->dx = -speed;
			za->dy = 0;
			za->n = (width - 10) * 256 / speed;
			break;
		case 2:
			za->w = width;
			za->h = 10;
			za->x = 0;
			za->y = height - 10;
			za->dx = 0;
			za->dy = -speed;
			za->n = (height - 10) * 256 / speed;
			break;
		case 3:
			za->w = 10;
			za->h = height;
			za->x = 0;
			za->y = 0;
			za->dx = speed;
			za->dy = 0;
			za->n = (width - 10) * 256 / speed;
			break;
		}
		za->ww = width - za->w;
		za->hh = height - za->h;

		/* We want smaller angle increments in sweep mode (looks better) */

		za->a1 = 0;
		za->a2 = 0;
		za->inc1 = ((2 * (random() & 1)) - 1) * (1 + random () % 7);
		za->inc2 = ((2 * (random() & 1)) - 1) * (1 + random () % 7);
	} else {
		za->w = 50 + random() % 300;
		za->h = 50 + random() % 300;

		if (za->w > width / 3)
			za->w = width / 3;
		if (za->h > height / 3)
			za->h = height / 3;

		za->ww = width - za->w;
		za->hh = height - za->h;

		za->x = (random() % za->ww);
		za->y = (random() % za->hh);

		za->dx = ((2 * (random() & 1)) - 1) * (100 + random() % 300);
		za->dy = ((2 * (random() & 1)) - 1) * (100 + random() % 300);

		if (anim) {
			za->n = 50 + random() % 1000;
			za->a1 = 0;
			za->a2 = 0;
		} else {
			za->n = 5 + random() % 10;
			za->a1 = random ();
			za->a2 = random ();
		}

		za->inc1 = ((2 * (random() & 1)) - 1) * (random () % 30);
		za->inc2 = ((2 * (random() & 1)) - 1) * (random () % 30);
	}

	za->xx = za->x * 256;
	za->yy = za->y * 256;

	za->count = 0;
}


static struct zoom_area *create_zoom (void)
{
	struct zoom_area *za;

	za = malloc (sizeof (struct zoom_area));
	reset_zoom (za);

	return za;
}


static void update_position (struct zoom_area *za)
{
	za->xx += za->dx;
	za->yy += za->dy;

	za->x = za->xx >> 8;
	za->y = za->yy >> 8;

	if (za->x < 0) {
		za->x = 0;
		za->dx = 100 + random() % 100;
	}
		
	if (za->y < 0) {
		za->y = 0;
		za->dy = 100 + random() % 100;
	}
		
	if (za->x > za->ww) {
		za->x = za->ww;
		za->dx = -(100 + random() % 100);
	}

	if (za->y > za->hh) {
		za->y = za->hh;
		za->dy = -(100 + random() % 100);
	}
}



static void hack_main (void)
{
	int i;

	for (i = 0; i < num_zoom; i++) {
		if (move || sweep)
			update_position (zoom_box[i]);

		if (zoom_box[i]->n > 0) {
			if (anim || zoom_box[i]->count == 0) {
				rotzoom (zoom_box[i]);
			} else {
				sleep (1);
			}
			zoom_box[i]->n--;
		} else {
			reset_zoom (zoom_box[i]);
		}
	}

	flushimage(display, 1);
	if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4) {
				if(emenuhit(3, &m, &menu) == 0) 
					exits(0);
			}
	}
}




void
eresized(int new)
{
	int i;

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window: %r");
	}

	width = Dx(screen->r);
	height = Dy(screen->r);

	if (width % 2)
		width--;
	if (height % 2)
		height--;

	if(scrimg != nil)
		freeimage(scrimg);
	scrimg = squeeze(scrbuf);
	if(scrimg == nil)
		sysfatal("unable to fit screenshot to window: %r");

	if(zoom_box == nil) {
		zoom_box = calloc (num_zoom, sizeof (struct zoom_area *));
		for(i = 0; i < num_zoom; i++)
			zoom_box[i] = create_zoom ();
	} else {		
		for (i = 0; i < num_zoom; i++) 
			reset_zoom(zoom_box[i]);
	}

	draw(screen, screen->r, display->black, nil, ZP);
}


void screenhack(void)
{

	/* In sweep or static mode, we want only one box */
	if (sweep || !anim)
		num_zoom = 1;

	/* Can't have static sweep mode */
	if (!anim)
		sweep = 0;

	/* Main drawing loop */
	while (42) {
		hack_main ();
		sleep (delay);
	}
}

void 
main(int argc, char **argv)
{

	USED(argc, argv);
	srand(time(0));

	num_zoom = 2;
	move = 0;
	delay = 0;
	sweep = 1;
	anim = 1;

	scrbuf = getscr();
	if(scrbuf == nil)
		sysfatal("could not obtain screenshot: %r");

	if(initdraw(nil, nil, "rotzoomer") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);
	eresized(0);

	screenhack();
}
